""" Обработка ошибок. """


class ServerError(Exception):
    """ Класс - исключение, для обработки ошибок сервера.
        При генерации требует строку с описанием ошибки,
        полученную с сервера. """

    def __init__(self, text):
        self.text = text

    def __str__(self):
        return self.text



""" Дескриптор. Инициализиция логера
метод определения модуля, источника запуска."""
import logging
import sys
import re

file = sys.argv[0]
if re.search(r'(server)', file):
    Logger = logging.getLogger('server')
else:
    Logger = logging.getLogger('client')


class Port:
    """ Класс - дескриптор для номера порта.
        Позволяет использовать только порты с 1023 по 65536. При попытке
        установить неподходящий номер порта генерирует исключение."""

    def __set__(self, instance, value):
        if not 1023 < value < 65536:
            Logger.critical(
                f'Попытка запуска с указанием неподходящего порта {value}.'
                f' Допустимы адреса с 1024 до 65535.')
            raise TypeError('Некорректрый номер порта')
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name



# """ Unit-Тесты утилит """
# import sys
# import os
# import unittest
# import json
#
# from packs.exe.server.common.utils import get_message
# from packs.pack_server.server.common.utils import send_message
# from packs.pack_server.server.common.variables import ENCODING, ACTION, PRESENCE, \
#     TIME, USER, ACCOUNT_NAME, RESPONSE, ERROR
#
# sys.path.append(os.path.join(os.getcwd(), '../../../..'))
#
#
# class TestSocket:
#     """ Тестовый класс для тестирования отправки и получения сообщения
#     при создании требует словарь, который будет прогоняться
#     через тестовую функцию. """
#
#     def __init__(self, test_dict):
#         self.test_dict = test_dict
#         self.encoded_message = None
#         self.received_message = None
#
#     def send(self, message_to_send):
#         """ Тестовая функция отправки данных, корректно кодирует сообщение,
#             так же сохраняет то, что должно было отправиться в сокет. """
#         json_test_message = json.dumps(self.test_dict)
#         self.encoded_message = json_test_message.encode(
#             ENCODING)  # Кодируем сообщение
#         # Сохраняем то, что должно было отправиться в сокет
#         self.received_message = message_to_send
#
#     def recv(self, max_len):
#         """ Получаем данные из сокета. """
#         json_test_message = json.dumps(self.test_dict)
#         return json_test_message.encode(ENCODING)
#
#
# """ Тестовый класс выполняющий тестирование. """
#
#
# class Tests(unittest.TestCase):
#     test_dict_send = {
#         ACTION: PRESENCE,
#         TIME: 111111.111111,
#         USER: {
#             ACCOUNT_NAME: "test_test"
#         }
#     }
#     test_dict_recv_ok = {RESPONSE: 200}
#     test_dict_recv_err = {
#         RESPONSE: 400,
#         ERROR: 'Bad Request'
#     }
#
#     """ Тестирование корректности работы функции отправки,для этого создаем тестовый сокет
#     и проверяем корректность отправки словаря. """
#
#     def test_send_message(self):
#         # экземпляр тестового словаря хранит - словарь.
#         test_socket = TestSocket(self.test_dict_send)
#         # вызов тестируемой функции, результат в тестовом сокете
#         send_message(test_socket, self.test_dict_send)
#         # Проверка корректности кодирования словаря.
#         self.assertEqual(
#             test_socket.encoded_message,
#             test_socket.received_message)
#         # Дополнительная проверка, если на входе не словарь
#         with self.assertRaises(Exception):
#             send_message(test_socket, test_socket)
#
#     """ Тест функции приема сообщения. """
#
#     def test_get_message(self):
#         test_sock_ok = TestSocket(self.test_dict_recv_ok)
#         test_sock_err = TestSocket(self.test_dict_recv_err)
#         # Тест коректной расшифровки коректного словаря
#         self.assertEqual(get_message(test_sock_ok), self.test_dict_recv_ok)
#         # Тест- корректная расшифровка ошибочного словаря ???
#         self.assertEqual(get_message(test_sock_err), self.test_dict_recv_err)
#
#
# if __name__ == '__main__':
#     unittest.main()



""" Основное окно сервера."""

from PyQt5.QtWidgets import QMainWindow, QAction, qApp, QLabel, QTableView
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtCore import QTimer

from stat_window import StatWindow
from config_window import ConfigWindow
from add_user import RegisterUser
from remove_user import DelUserDialog


class MainWindow(QMainWindow):
    """Класс - основное окно сервера."""

    def __init__(self, database, server, config):
        # Конструктор предка
        super().__init__()

        # База данных сервера.
        self.database = database

        self.server_thread = server
        self.config = config

        # Ярлык выхода.
        self.exitAction = QAction('Выход', self)
        self.exitAction.setShortcut('Ctrl+Q')
        self.exitAction.triggered.connect(qApp.quit)

        # Кнопка обновить список клиентов.
        self.refresh_button = QAction('Обновить список', self)

        # Кнопка настроек сервера.
        self.config_btn = QAction('Настройки сервера', self)

        # Кнопка регистрации пользователя.
        self.register_btn = QAction('Регистрация пользователя', self)

        # Кнопка удаления пользователя.
        self.remove_btn = QAction('Удаление пользователя', self)

        # Кнопка вывести историю сообщений.
        self.show_history_button = QAction('История клиентов', self)

        # Статусбар
        self.statusBar()
        self.statusBar().showMessage('Server Working')

        # Тулбар
        self.toolbar = self.addToolBar('MainBar')
        self.toolbar.addAction(self.exitAction)
        self.toolbar.addAction(self.refresh_button)
        self.toolbar.addAction(self.show_history_button)
        self.toolbar.addAction(self.config_btn)
        self.toolbar.addAction(self.register_btn)
        self.toolbar.addAction(self.remove_btn)

        # Настройки геометрии основного окна
        # Поскольку работать с динамическими размерами мы не умеем, и мало
        # времени на изучение, размер окна фиксирован.
        self.setFixedSize(800, 600)
        self.setWindowTitle('Messaging Server alpha release')

        # Надпись о том, что ниже список подключённых клиентов
        self.label = QLabel('Список подключённых клиентов:', self)
        self.label.setFixedSize(240, 15)
        self.label.move(10, 25)

        # Окно со списком подключённых клиентов.
        self.active_clients_table = QTableView(self)
        self.active_clients_table.move(10, 45)
        self.active_clients_table.setFixedSize(780, 400)

        # Таймер, обновляющий список клиентов 1 раз в секунду
        self.timer = QTimer()
        self.timer.timeout.connect(self.create_users_model)
        self.timer.start(1000)

        # Связываем кнопки с процедурами
        self.refresh_button.triggered.connect(self.create_users_model)
        self.show_history_button.triggered.connect(self.show_statistics)
        self.config_btn.triggered.connect(self.server_config)
        self.register_btn.triggered.connect(self.reg_user)
        self.remove_btn.triggered.connect(self.rem_user)

        # Последним параметром отображаем окно.
        self.show()

    def create_users_model(self):
        """Метод заполняющий таблицу активных пользователей."""
        list_users = self.database.active_users_list()
        list_active_users = QStandardItemModel()
        list_active_users.setHorizontalHeaderLabels(
            ['Имя Клиента', 'IP Адрес', 'Порт', 'Время подключения'])
        for row in list_users:
            user, ip_address, port, time = row
            user = QStandardItem(user)
            user.setEditable(False)
            ip_address = QStandardItem(ip_address)
            ip_address.setEditable(False)
            port = QStandardItem(str(port))
            port.setEditable(False)
            # Уберём милисекунды из строки времени, т.к. такая точность не
            # требуется.
            time = QStandardItem(str(time.replace(microsecond=0)))
            time.setEditable(False)
            list_active_users.appendRow([user, ip_address, port, time])
        self.active_clients_table.setModel(list_active_users)
        self.active_clients_table.resizeColumnsToContents()
        self.active_clients_table.resizeRowsToContents()

    def show_statistics(self):
        """Метод создающий окно со статистикой клиентов."""
        global stat_window
        stat_window = StatWindow(self.database)
        stat_window.show()

    def server_config(self):
        """Метод создающий окно с настройками сервера."""
        global config_window
        # Создаём окно и заносим в него текущие параметры
        config_window = ConfigWindow(self.config)

    def reg_user(self):
        """Метод создающий окно регистрации пользователя."""
        global reg_window
        reg_window = RegisterUser(self.database, self.server_thread)
        reg_window.show()

    def rem_user(self):
        """Метод создающий окно удаления пользователя."""
        global rem_window
        rem_window = DelUserDialog(self.database, self.server_thread)
        rem_window.show()



""" Класс - диалог выбора контакта для удаления."""
from PyQt5.QtWidgets import QDialog, QLabel, QComboBox, QPushButton
from PyQt5.QtCore import Qt


class DelUserDialog(QDialog):
    def __init__(self, database, server):
        super().__init__()
        self.database = database
        self.server = server

        self.setFixedSize(350, 120)
        self.setWindowTitle('Удаление пользователя')
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(True)

        self.selector_label = QLabel(
            'Выберите пользователя для удаления:', self)
        self.selector_label.setFixedSize(200, 20)
        self.selector_label.move(10, 0)

        self.selector = QComboBox(self)
        self.selector.setFixedSize(200, 20)
        self.selector.move(10, 30)

        self.btn_ok = QPushButton('Удалить', self)
        self.btn_ok.setFixedSize(100, 30)
        self.btn_ok.move(230, 20)
        self.btn_ok.clicked.connect(self.remove_user)

        self.btn_cancel = QPushButton('Отмена', self)
        self.btn_cancel.setFixedSize(100, 30)
        self.btn_cancel.move(230, 60)
        self.btn_cancel.clicked.connect(QDialog.close)
        # self.btn_cancel.clicked.connect(self.close)

        self.all_users_fill()

    def all_users_fill(self):
        """Метод заполняющий список пользователей."""
        self.selector.addItems([item[0]
                                for item in self.database.users_list()])

    def remove_user(self):
        """Метод - обработчик удаления пользователя."""
        self.database.remove_user(self.selector.currentText())
        if self.selector.currentText() in self.server.names:
            sock = self.server.names[self.selector.currentText()]
            del self.server.names[self.selector.currentText()]
            self.server.remove_client(sock)
        # Рассылаем клиентам сообщение о необходимости обновить справочники
        self.server.service_update_lists()
        self.close()



""" Декораторы. """

import socket
import logging
import sys

sys.path.append('..')

# метод определения модуля, источника запуска.
if sys.argv[0].find('client') == -1:
    # если не клиент, то сервер!
    logger = logging.getLogger('server')
else:
    # иначе сервер
    logger = logging.getLogger('client')


def log(func_to_log):
    """ Декоратор, выполняющий логирование вызовов функций.
        Сохраняет события типа debug, содержащие
        информацию об имени вызываемой функиции, параметры с которыми
        вызывается функция, и модуль, вызывающий функцию."""

    def log_saver(*args, **kwargs):
        logger.debug(
            f'Была вызвана функция {func_to_log.__name__} c параметрами '
            f'{args}, {kwargs}. Вызов из модуля {func_to_log.__module__}')
        ret = func_to_log(*args, **kwargs)
        return ret

    return log_saver


def login_required(func):
    """ Декоратор, проверяющий, что клиент авторизован на сервере.
        Проверяет, что передаваемый объект сокета находится в
        списке авторизованных клиентов.
        За исключением передачи словаря-запроса
        на авторизацию. Если клиент не авторизован,
        генерирует исключение TypeError"""

    def checker(*args, **kwargs):
        # проверяем, что первый аргумент - экземпляр MessageProcessor
        # Импортить необходимо тут, иначе ошибка рекурсивного импорта.
        from packs.pack_server.server.server.core import MessageProcessor
        from variables import ACTION, PRESENCE
        if isinstance(args[0], MessageProcessor):
            found = False
            for arg in args:
                if isinstance(arg, socket.socket):
                    # Проверяем, что данный сокет есть в списке names класса
                    # MessageProcessor
                    for client in args[0].names:
                        if args[0].names[client] == arg:
                            found = True

            # Теперь надо проверить, что передаваемые аргументы не presence
            # сообщение. Если presence, то разрешаем
            for arg in args:
                if isinstance(arg, dict):
                    if ACTION in arg and arg[ACTION] == PRESENCE:
                        found = True
            # Если не авторизован и не сообщение начала авторизации, то
            # вызываем исключение.
            if not found:
                raise TypeError
        return func(*args, **kwargs)

    return checker



""" Config серверного логгера. """

from packs.pack_server.server.common.variables import LOGGING_LEVEL
import logging.handlers
import os
import sys

sys.path.append('../')


# Создание формировщика логов.
SERVER_FORMATTER = logging.Formatter(
    '%(asctime)s %(levelname)s %(filename)s %(message)s')

# Подготовка имени файла для логирования.
# PATH = os.path.dirname(os.path.abspath(__file__))
PATH = os.getcwd()
PATH = os.path.join(PATH, 'server.log')

# Создание потоков вывода логов.
STREAM_HANDLER = logging.StreamHandler(sys.stderr)
STREAM_HANDLER.setFormatter(SERVER_FORMATTER)
STREAM_HANDLER.setLevel(logging.ERROR)
LOG_FILE = logging.handlers.TimedRotatingFileHandler(
    PATH, encoding='utf-8', interval=1, when='D')
LOG_FILE.setFormatter(SERVER_FORMATTER)

# Создание и настройка регистратора.
LOGGER = logging.getLogger('server')
LOGGER.addHandler(STREAM_HANDLER)
LOGGER.addHandler(LOG_FILE)
LOGGER.setLevel(LOGGING_LEVEL)

# Отладка.
if __name__ == '__main__':
    LOGGER.critical('Критическая ошибка!!!')
    LOGGER.error('Ошибка!')
    LOGGER.warning('Предупреждение!')
    LOGGER.debug('Отладочная информация.')
    LOGGER.info('Информационное сообщение.')



from setuptools import setup, find_packages

setup(name="messenger_server_from_nadezhda",
      version="0.3.0",
      description="Server packet",
      author="Nadezhda Vlasova",
      author_email="nadvlasova@rambler.ru",
      packages=find_packages(),
      install_requires=['PyQt5', 'sqlalchemy', 'pycryptodome', 'pycryptodomex']
      )



""" Функции отправки и приема сообщений."""
import json
import sys

from variables import MAX_PACKAGE_LENGTH, ENCODING

sys.path.append('../')


# @log
def get_message(client):
    """
    Функция приёма сообщений от удалённых компьютеров.
    Принимает сообщения JSON, декодирует полученное сообщение
    и проверяет что получен словарь.
    :param client: сокет для передачи данных.
    :return: словарь - сообщение.
    """
    encoded_response = client.recv(MAX_PACKAGE_LENGTH)
    json_response = encoded_response.decode(ENCODING)
    response = json.loads(json_response)
    if isinstance(response, dict):
        return response
    raise TypeError


# @log
def send_message(sock, message):
    """
    Функция отправки словарей через сокет.
    Кодирует словарь в формат JSON и отправляет через сокет.
    :param sock: сокет для передачи
    :param message: словарь для передачи
    :return: ничего не возвращает
    """
    js_message = json.dumps(message)
    encoded_message = js_message.encode(ENCODING)
    sock.send(encoded_message)



# """ Тестирование сервера. """
# import unittest
#
# from common.variables import RESPONSE, ERROR, TIME, USER, ACCOUNT_NAME, \
#     ACTION, PRESENCE
# from server import process_client_message
#
#
# class TestServer(unittest.TestCase):
#     def setUp(self):
#         pass
#
#     err_dict = {
#         RESPONSE: 400,
#         ERROR: 'Bad Request'
#     }
#     ok_dict = {RESPONSE: 200}
#
#     # Тест-ошибка, если нет действия
#     def test_no_action(self):
#         self.assertEqual(process_client_message(
#             {TIME: '1.1', USER: {ACCOUNT_NAME: 'Guest'}}), self.err_dict)
#
#     # Тест-ошибка, если действие не корректное
#     def test_wrong_action(self):
#         self.assertEqual(process_client_message(
#             {ACTION: 'Wrong', TIME: '1.1', USER: {ACCOUNT_NAME: 'Guest'}}),
#             self.err_dict)
#
#     # Тест на наличие штампа времени
#     def test_no_time(self):
#         self.assertEqual(process_client_message(
#             {ACTION: PRESENCE, USER: {ACCOUNT_NAME: 'Guest'}}), self.err_dict)
#
#     # Тест на наличие пользователя
#     def test_no_user(self):
#         self.assertEqual(process_client_message(
#             {ACTION: PRESENCE, TIME: '1.1'}), self.err_dict)
#
#     # Тест неизвестный пользователь
#     def test_unknown_user(self):
#         self.assertEqual(process_client_message(
#             {ACTION: PRESENCE, TIME: 1.1,
#              USER: {ACCOUNT_NAME: 'Guest1'}}),
#             self.err_dict)
#
#     # Тест на корректный запрос
#     def test_ok_check(self):
#         self.assertEqual(process_client_message(
#             {ACTION: PRESENCE, TIME: 1.1, USER: {ACCOUNT_NAME: 'Guest'}}),
#             self.ok_dict)
#
#     def tearDown(self):
#         pass
#
#
# if __name__ == '__main__':
#     unittest.main()

# разобраться почему где-то TIME: 1.1 а где-то TIME: '1.1' ???????????????
